import os
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from . import palettes
import errno

def colors_from_hue(data, hue, cmap):
    num_colors = len(data.groupby(hue))
    cm_subsection = np.linspace(0.2, 0.8, num_colors+2)
    cm_subsection = cm_subsection[1:num_colors+1]
    return [ cmap(v) for v in cm_subsection ]

def hueize(data, hue=None, cmap=palettes.neon, *args, **kwargs):
    """
    Groups a dataframe by hues, if any, and generates plot settings for each hue.

    Returns
    -------
      Generator of (pd.DataFrame, kwargs), where the kwargs can be used in matplotlib functions.
    
    Parameters
    ----------

      data: pandas.DataFrame
        dataframe to plot

      x: string
        column of data to plot

    Keyword Arguments
    -----------------

      hue: string
        column of dataframe to split on
    """        
    if hue is None:
        new_kwargs = kwargs.copy()
        new_kwargs.update({'color': kwargs.get('color', cmap(0.5))})
        yield (data, new_kwargs)
    else:
        if cmap:
            colors = colors_from_hue(data, hue, cmap)
        elif 'colors' in kwargs:
            colors = kwargs['colors']
            del kwargs['colors']
        else:
            colors = [kwargs.get('color')]
        
        for i,(label,grp) in enumerate(data.groupby(hue)):
            new_kwargs = kwargs.copy()

            # matplotlib doesn't like it when we have these in its kwargs.
            new_kwargs.pop('markers', None)
            new_kwargs.pop('linestyles', None)
            
            if 'markers' in kwargs:
                markers = kwargs['markers']
                new_kwargs['marker'] = markers[i % len(markers)]

            if 'linestyles' in kwargs:
                linestyles = kwargs['linestyles']
                new_kwargs['linestyle'] = linestyles[i % len(linestyles)]

            new_kwargs.update({
                'label': label,
                'color': colors[i % len(colors)],
            })
            
            yield (grp, new_kwargs)


def plot(data, x, y, error=None, *args, **kwargs):
    data = data.sort_values(x, ascending=True)

    for df,kwargs in hueize(data, *args, **kwargs):
        if error is not None:
            plt.errorbar(list(df[x]), list(df[y]), yerr=df[error], **kwargs)
        else:
            plt.plot(list(df[x]), list(df[y]), **kwargs)

    plt.ylabel(y)
    plt.xlabel(x)
    plt.gca().autoscale(tight=True)
    plt.gca().margins(y=0.1)

def stackplot(data, x, y, hue, cmap=palettes.neon):
    xs = np.array(data[x])
    yss = []
    labels = []
    for k,df in data.groupby(hue):
        labels.append(k)
        df_xs = grp[x].tolist()
        grp_ys = grp[y].tolist()
        ys = []
        for v in xs:
            if len(grp_xs) > 0 and grp_xs[0] == v:
                ys.append(grp_ys.pop(0))
                grp_xs.pop(0)
            else:
                if len(ys) == 0:
                    ys.append(0.)
                else:
                    ys.append(ys[-1])
        assert len(grp_xs) == 0
        assert len(grp_ys) == 0
        assert len(ys) == len(xs)
        yss.append(np.array(ys, dtype=float))

    if cmap is not None:
        colors = colors_from_hue(data, hue, cmap)
    else:
        colors = None

    plt.stackplot(xs, *yss, labels=labels, colors=colors)

    plt.ylabel(y)
    plt.xlabel(x)
    plt.gca().autoscale(tight=True)
    plt.gca().margins(y=0.1)

def scatter(data, x, y, markers=['o'], linestyles=[''], **kwargs):
    return plot(data=data, x=x, y=y, markers=markers, linestyles=linestyles, **kwargs)

def hist(data, x, alpha=0.5, rwidth=0.92, *args, **kwargs):
    """
    Plot a histogram from a dataframe column.

    If hue is provided, plot many overlayed histograms, one per value of the data[hue] column.

    Parameters
    ----------

      data: pandas.DataFrame
        dataframe to plot

      x: string
        column of data to plot

    Keyword Arguments
    -----------------

      hue: string
        column of dataframe to split on

      alpha: float
        opacity of histogram (default: 0.5)

      rwidth: float
        The relative width of the bars as a fraction of the bin width (default: 0.92)
    """
    new_kwargs = dict(alpha=alpha, rwidth=rwidth)
    new_kwargs.update(kwargs)
    for df,kwargs in hueize(data, *args, **new_kwargs):
        plt.hist(df[x], **kwargs)
    plt.xlabel(x)
    plt.ylabel("Frequency")
        
def cdf(data, x, *args, **kwargs):
    """
    Plot a cdf from a dataframe column.

    If hue is provided, plot many overlayed cdfs, one per value of the data[hue] column.

    Parameters
    ----------

      data: pandas.DataFrame
        dataframe to plot

      x: string
        column of data to plot

    Keyword Arguments
    -----------------

      hue: string
        column of dataframe to split on
    """    
    for df,kwargs in hueize(data, *args, **kwargs):
        sorted_data = np.sort(df[x])
        sorted_data_cdf = np.arange(len(sorted_data))/float(len(sorted_data)) * 100 
        plt.plot(sorted_data, sorted_data_cdf, *args, **kwargs)
    plt.ylabel("Cumulative % Data")
    plt.xlabel(x)
    
def pdf(data, x, bins=10, normalize=True, *args, **kwargs):
    """
    Plot a probability density function (pdf) from a dataframe column.

    If hue is provided, plot many overlayed pdf, one per value of the data[hue] column.

    Parameters
    ----------

      data: pandas.DataFrame
        dataframe to plot

      x: string
        column of data to plot

    Keyword Arguments
    -----------------

      hue: string
        column of dataframe to split on

      bins: int or sequence of scalars or str, optional
        Bins to use for the function. See `numpy.histogram <https://numpy.org/doc/stable/reference/generated/numpy.histogram.html>`

      normalize: boolean
        Normalize to a probability density (default). If false, returns a histogram.
    """    
    for df,kwargs in hueize(data, *args, **kwargs):
        ys,xs = np.histogram(df[x], bins=bins)
        if normalize:
            ys = np.array(ys)/sum(ys)*100
        plt.plot(xs[:-1], ys, *args, **kwargs)
    plt.ylabel("% Data")
    plt.xlabel(x)
    
def savefig(filename, **kwargs):
    """Saves a figure, but also creates the directory and calls tight_layout before saving"""
    if not os.path.exists(os.path.dirname(filename)):
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError as exc: # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise

    if 'bbox_inches' not in kwargs:
        kwargs['bbox_inches'] = 'tight'
    if 'pad_inches' not in kwargs:
        kwargs['pad_inches'] = 0
            
    plt.tight_layout(pad=0)
    plt.savefig(filename, **kwargs)



# Most of this is lifted from Seaborn (https://stanford.edu/~mwaskom/software/seaborn/)
#
# Copyright (c) 2012-2013, Michael L. Waskom
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the {organization} nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import division
import matplotlib
import matplotlib.pyplot as plt

# Does a bunch of styling stuff that needs to be done manually instead of in the
# rcparams
def style_axis(show_xaxis=False, label_fontweight='medium', label_fontstyle='italic'):
    ax = plt.gca()

    ax.xaxis.grid(show_xaxis)    
    ax.xaxis.label.set_fontweight(label_fontweight)
    ax.xaxis.label.set_fontstyle(label_fontstyle)
    
    # position the y label at the top left, left-aligned with the ticks.
    left_in_axis = [0, 1]
    left_in_display = ax.transAxes.transform(left_in_axis)
    left_label_in_display = [left_in_display[0] - ax.yaxis.get_tick_space() * ax.figure.dpi/72.,
                             left_in_display[1] + ax.xaxis.get_tick_space() * ax.figure.dpi/72.]
    left_label_in_axis = ax.transAxes.inverted().transform(left_label_in_display)

    ax.yaxis.set_label_coords(*left_label_in_axis)
    ax.yaxis.label.set_fontweight(label_fontweight)
    ax.yaxis.label.set_fontstyle(label_fontstyle)
    ax.yaxis.label.set_ha('left')
    ax.yaxis.label.set_va('bottom')
    ax.yaxis.label.set_rotation(0)

    # position the y label at the top left, left-aligned with the ticks.
    title_in_display = [left_in_display[0] - ax.yaxis.get_tick_space() * ax.figure.dpi/72.,
                        left_label_in_display[1] + ax.xaxis.get_tick_space() * ax.figure.dpi/72.]
    title_in_axis = ax.transAxes.inverted().transform(title_in_display)
    ax.title.set_position(title_in_axis)
    ax.title.set_fontweight("bold")
    ax.title.set_ha("left")
    ax.title.set_ha("left")
    
    plt.tight_layout()



from __future__ import division

from . import palettes
from .style import *
from .plotting import *
from .legend import *

import matplotlib as mpl
import os
dirname = os.path.dirname(os.path.abspath(__file__))
mpl.style.core.USER_LIBRARY_PATHS.append(dirname)
mpl.style.core.reload_library()
